#! /usr/bin/env python

from __future__ import with_statement

import os
import platform
import subprocess
import sys
import shutil
from contextlib import contextmanager
import xml.etree.cElementTree as etree


PYTHON_VERSION = ".".join((str(sys.version_info[0]), str(sys.version_info[1])))
OS_VERSION = tuple(int(v) for v in platform.mac_ver()[0].split('.'))

@contextmanager
def workingdirectory(directory):
    wd = os.getcwd()
    os.chdir(directory)
    yield directory
    os.chdir(wd)

def bool_(v):
    return str(bool(v)).lowercase()

class PackageMakerDocument(object):

    FORMAT_VERSION = '1.12'

    @classmethod
    def load(cls, document_path):
        return cls(document_path)

    def __init__(self, document_path=None):
        self.document_path = document_path
        if document_path:
            self.root = etree.parse(os.path.join(document_path, 'index.xml'))
            self.properties = self.root.find('properties')
            self.distribution = self.root.find('distribution')
            self.contents = self.root.find('contents')
            self.resources = self.root.find('resources')
            self.requirements = self.root.find('requirements')
            self.flags = self.root.find('flags')
        else:
            self.root = etree.Element('pkmkdoc', spec=self.FORMAT_VERSION)
            self.properties = etree.SubElement(self.root, 'properties')
            self.distribution = etree.SubElement(self.root, 'distribution')
            self.contents = etree.SubElement(self.root, 'contents')
            self.resources = etree.SubElement(self.root, 'resources')
            self.requirements = etree.SubElement(self.root, 'requirements')
            self.flags = etree.SubElement(self.root, 'flags')

    def add_choice(self, title, id, pkgid, starts_selected=True, starts_enabled=True, starts_hidden=False):
        choice = etree.SubElement(self.contents, 'choice', title=title, id=id,
                                  starts_selected=bool_(starts_selected),
                                  starts_enabled=bool_(starts_enabled),
                                  starts_hidden=bool_(starts_hidden))
        etree.SubElement(choice, 'pkgref', id=pkgid)

    def get_choice(self, title=None, id=None):
        for choice in root.iter('choice'):
            if (choice.get('title') == title or not title) and (choice.get('id') == id or not id):
                return choice

class Builder(object):

    if OS_VERSION[1] < 4:
        raise Exception("This version of Mac OS X is not supported because it is too old.")
    elif OS_VERSION[1] == 4:
        CC = 'gcc-4.0'
    elif OS_VERSION[1] == 5:
        CC = 'gcc-4.0'
    elif OS_VERSION[1] == 6:
        CC = 'llvm-gcc-4.2'
    else:
        raise Exception("Unknown version of Mac OS X.")

    def __init__(self, pmdoc='Bazaar.pmdoc', desktop=False):
        self.pmdoc = pmdoc
        self.desktop = desktop
        self.root = os.path.dirname(os.path.abspath(sys.argv[0]))
        self.dstroot = None
        self.srcroot = os.path.join(self.root, 'bazaar.src')
        self.pkgroot = os.path.join(self.root, 'package')
        if OS_VERSION[1] == 4:
            self.sitepackages_prefix = 'Library/Frameworks/Python.framework/Versions/%s/lib/python%s/site-packages' % \
                (sys.version[:3], sys.version[:3])
        else:
            self.sitepackages_prefix = os.path.join('Library/Python', sys.version[:3], 'site-packages')

# FS utilities

    @contextmanager
    def dstrootdirectory(self, directory):
        old_dstroot = self.dstroot
        self.dstroot = os.path.join(self.root, 'package', os.path.basename(directory) + '.dst')
        if not os.path.isdir(self.dstroot):
            os.mkdir(self.dstroot)
        yield self.dstroot
        self.dstroot = old_dstroot

    def subprocess(self, args, env_update=None, **kwargs):
        if 'env' not in kwargs:
            kwargs['env'] = dict(os.environ)
        if env_update:
            kwargs['env'].update(env_update)
        if 'CC' not in kwargs['env']:
            kwargs['env']['CC'] = self.CC
        if 'shell' not in kwargs:
            kwargs['shell'] = False
        try:
            subprocess.check_call(args, **kwargs)
        except subprocess.CalledProcessError, e:
            print 'Command failed with non-zero exit status %d' % e.returncode
            print '    command: ' + str(args)
            print '    env: ' + '\n         '.join('%s: %s' % (k, v) for k, v in kwargs['env'].iteritems())
            sys.exit(1)

    def python(self, args, **kwargs):
        args = [sys.executable] + list(args)
        self.subprocess(args, **kwargs)

    def ditto(self, src, dst):
        src = os.path.abspath(src)
        if os.path.isdir(dst):
            dst = os.path.join(os.path.abspath(dst), os.path.basename(src))
        else:
            dst = os.path.abspath(dst)
        self.subprocess(['ditto', src, dst], shell=False)
        return dst

    def dstditto(self, src, dst):
        assert(self.dstroot)
        return self.ditto(src, os.path.normpath(self.dstroot + '/' + dst))

    def rm(self, src):
        # use rm to be able to delete both files and folders
        self.subprocess(['rm', '-rf', os.path.abspath(src)])
        
# building

    def build_package(self, src):
        with workingdirectory(src):
            with self.dstrootdirectory(src):
                self.python(['setup.py', 'install', '--root=' + self.dstroot])

    def build_package_bzr(self, src):
        with workingdirectory(src):
            with self.dstrootdirectory(src):
                self.build_package(src)
                self.rm('System')
                self.dstditto('bzr.1', 'usr/local/share/man/man1/bzr.1')
                self.ditto('COPYING.txt', self.pkgroot)
        self.bzrlib_path = os.path.join(self.sitepackages_prefix, 'bzrlib')

    def build_package_sip(self, src):
        with workingdirectory(src):
            with self.dstrootdirectory(src):
                config = ['-b', os.path.join(self.dstroot, 'usr/local/bin'), 
                  '-d', os.path.join(self.dstroot, self.sitepackages_prefix), 
                  '-e', os.path.join(self.dstroot, 'usr/local/include/python' + PYTHON_VERSION),
                  '-v', os.path.join(self.dstroot, 'usr/local/share/sip')]
                self.python(['configure.py'] + config)
                self.subprocess(['make'])
                self.subprocess(['make', 'install'])

    def build_package_pyqt(self, src):
        with workingdirectory(src):
            with self.dstrootdirectory(src):
                config = [
                    '-q', '/usr/bin/qmake-4.5',
                    '-d', os.path.join(self.dstroot, self.sitepackages_prefix),
                    '-b', os.path.join(self.dstroot, 'usr/local/bin'),
                    '--no-sip-files',
                    '--confirm-license',
                    '--no-designer-plugin'
                ]
                self.python(['configure.py'] + config)
                self.subprocess(['make'])
                self.subprocess(['make', 'install'])

    def build_plugin(self, src, dst_dir):
        with self.dstrootdirectory(src):
            os.makedirs(os.path.join(self.dstroot, self.bzrlib_path))
            plugin_dst = self.dstditto(src, os.path.join(self.bzrlib_path, 'plugins', dst_dir))
            with workingdirectory(plugin_dst):
                self.python(['setup.py', 'build_ext', '--inplace'])
        self.rm(os.path.join(plugin_dst, 'build'))

    def build_plugin_bzrtools(self, src, dst_dir):
        with workingdirectory(src):
            self.rm('.be')
        self.build_plugin(src, dst_dir)

    def build_distribution(self):
        print "building distribution"
        with self.dstrootdirectory('bzr'):
            sys.path.insert(0, os.path.join(self.dstroot, self.sitepackages_prefix))
            bzrlib = __import__('bzrlib', globals(), None, [], 0)
        args = ['/Developer/Applications/Utilities/PackageMaker.app/Contents/MacOS/PackageMaker']
        pkg_filename = 'Bazaar%s-%s.pkg' % (bzrlib.version_string, config.DISTRIBUTION_VERSION)
        if OS_VERSION[1] == 4:
            args.extend(['build', '--proj', os.path.join(self.root, 
                         'Bazaar-%s.pmproj' % sys.version[:3]), '-p', 
                         pkg_filename])
        else:
            args.extend(['--doc', os.path.join(self.root, self.pmdoc),
                         '--out', pkg_filename])
        self.subprocess(args)

    def clean_dstroots(self):
        with workingdirectory(self.pkgroot):
            for item in os.listdir(self.pkgroot):
                if item.endswith('.dst') and os.path.isdir(item):
                    self.rm(item)

    def build_packages(self):
        packages = config.packages
        if self.desktop:
            packages.extend(config.desktop_packages)
        for package in packages:
            source = config.extract(os.path.join(self.root, 'src',
                                                 config.get_tarfile(package)),
                                    package['destdir'])
            src = self.ditto(source, self.srcroot)
            getattr(self, 'build_package_' + package['destdir'], self.build_package)(src)
            self.rm(os.path.dirname(source))

    def build_plugins(self):
        plugins = config.plugins
        if self.desktop:
            plugins.extend(config.desktop_plugins)
        for plugin in plugins:
            source = config.extract(os.path.join(self.root, 'src', 
                                                 config.get_tarfile(plugin)),
                                    plugin['destdir'])
            src = self.ditto(source, self.srcroot)
            getattr(self, 'build_plugin_' + plugin['destdir'],
                    self.build_plugin)(src, plugin['destdir'])
            self.rm(os.path.dirname(source))

    def main(self):
        self.rm(self.srcroot)
        os.mkdir(self.srcroot)
        self.clean_dstroots()
        self.build_packages()
        self.build_plugins()
        self.build_distribution()
        self.rm(self.srcroot)

if __name__ == '__main__':
    import optparse
    parser = optparse.OptionParser()
    parser.add_option('--pmdoc', dest='pmdoc', default='Bazaar.pmdoc')
    parser.add_option('-d', '--desktop', dest='desktop', action='store_true', default=False)
    parser.add_option('-c', '--config', dest='config', action='store', default='config')
    (options, args) = parser.parse_args()
    
    try:
        global config
        config = __import__(options.config, globals(), locals(), [], 0)
    except ImportError:
        print "ERROR: failed to import config file"
        sys.exit(1)
    try:
        builder = Builder(options.pmdoc, options.desktop)
        builder.main()
    except KeyboardInterrupt:
        sys.exit(1)
